use crate::{
    consts::{
        BRAINPOOL_MIDDLE,
        BRAINPOOL_STRONG,
        BRAINPOOL_WEAK,
        SECP_MIDDLE,
        SECP_STRONG,
        SECP_WEAK,
    },
    macros::DefDer,
    match_error,
    structs::Error,
};
use log::*;
use macro_rules_attribute::macro_rules_derive;
use openssl::pkey::{
    Id,
    PKey,
    Private,
};
use serde::{
    Deserialize,
    de::DeserializeOwned,
};
use std::{
    collections::HashMap,
    net::IpAddr,
};
use tokio::{
    fs::File,
    io::AsyncReadExt,
};

type VString = Vec<String>;

#[macro_rules_derive(DefDer)]
#[derive(Deserialize)]
pub struct GeneralConfig {
    #[serde(default = "GeneralConfig::default_accounts")]
    pub accounts_path: String,
    #[serde(default = "GeneralConfig::default_sites")]
    pub sites_path: String,
    #[serde(default = "GeneralConfig::default_challenge")]
    pub http_challenge_path: Option<String>,
    #[serde(default = "GeneralConfig::default_dns")]
    pub dns: Option<Dns>,
    #[serde(default = "GeneralConfig::default_certificates")]
    pub certificates_path: String,
    #[serde(default = "GeneralConfig::default_cas")]
    pub ca: HashMap<String, CA>,
}

impl Default for GeneralConfig {
    #[inline]
    fn default() -> Self {
        Self {
            accounts_path: Self::default_accounts(),
            sites_path: Self::default_sites(),
            http_challenge_path: Self::default_challenge(),
            dns: Self::default_dns(),
            certificates_path: Self::default_certificates(),
            ca: Self::default_cas(),
        }
    }
}

impl GeneralConfig {
    #[inline]
    fn default_accounts() -> String {
        "accounts".into()
    }

    #[inline]
    fn default_sites() -> String {
        "sites".into()
    }

    #[inline]
    fn default_challenge() -> Option<String> {
        None
    }

    #[inline]
    fn default_dns() -> Option<Dns> {
        None
    }

    #[inline]
    fn default_cas() -> HashMap<String, CA> {
        HashMap::new()
    }

    #[inline]
    fn default_certificates() -> String {
        "certificates".into()
    }
}

#[macro_rules_derive(DefDer)]
#[derive(Deserialize)]
pub struct Dns;

#[macro_rules_derive(DefDer)]
#[derive(Deserialize)]
pub struct Eab {
    #[serde(rename = "eab_token", alias = "id")]
    pub token: String,
    #[serde(rename = "eab_key", alias = "key")]
    pub key: String,
}

impl Eab {
    pub fn key(&self) -> Result<PKey<Private>, Error> {
        let decoded = &match_error!(data_encoding::BASE64URL_NOPAD.decode(self.key.as_bytes())=>Err(error)-> "Failed to decode the HMAC key for the eab_key: {error}", Error::err("Failed to decode eab_key".into()));
        PKey::hmac(decoded).map_err(|error| Error::new(format!("Failed to parse the private key: {error}")))
    }
}

#[macro_rules_derive(DefDer)]
#[derive(Deserialize)]
pub struct CA {
    /// Url for the directory
    pub directory: String,

    /// Email addresses for the CA to contact the user
    pub email_addresses: Option<VString>,

    #[serde(flatten, default)]
    pub eab: Option<Eab>,

    /// Amount of days the certificate is renewed before the Certificate is outdated
    /// TODO: give to processor
    #[serde(default = "CA::default_renew")]
    pub renew_before: u32,

    #[serde(default)]
    pub tos_accepted: bool,
}

impl CA {
    fn default_renew() -> u32 {
        7
    }
}

#[macro_rules_derive(DefDer)]
#[derive(Copy, Deserialize, Default)]
pub enum Algorithm {
    Rsa,
    Brainpool,
    Secp,
    #[default]
    ED25519,
}

#[macro_rules_derive(DefDer)]
#[derive(Copy, Deserialize, Default)]
pub enum Strength {
    Weak,
    Middle,
    #[default]
    Strong,
}

impl Strength {
    pub fn rsabits(self) -> u32 {
        self as u32
    }
}

pub fn match_algo(key: &PKey<Private>, algorithm: Algorithm, strength: Strength) -> bool {
    match (key.id(), algorithm) {
        (Id::ED25519, Algorithm::ED25519) => true,
        (Id::RSA, Algorithm::Rsa) if key.bits() == strength.rsabits() => true,
        (Id::EC, Algorithm::Secp) | (Id::EC, Algorithm::Brainpool) => {
            let pkey = key.ec_key().unwrap();
            let curve = pkey.group().curve_name().unwrap();
            match (algorithm, strength) {
                (Algorithm::Secp, Strength::Weak) if SECP_WEAK == curve => true,
                (Algorithm::Secp, Strength::Middle) if SECP_MIDDLE == curve => true,
                (Algorithm::Secp, Strength::Strong) if SECP_STRONG == curve => true,
                (Algorithm::Brainpool, Strength::Weak) if BRAINPOOL_WEAK == curve => true,
                (Algorithm::Brainpool, Strength::Middle) if BRAINPOOL_MIDDLE == curve => true,
                (Algorithm::Brainpool, Strength::Strong) if BRAINPOOL_STRONG == curve => true,
                _ => false,
            }
        },
        _ => false,
    }
}

#[macro_rules_derive(DefDer)]
#[derive(Deserialize, Default)]
pub struct SiteConfig {
    /// The Configured Certificate Authority
    pub ca: String,

    /// The Domains this site is responsible for
    pub domains: VString,

    /// IPAddresses for the Certificate
    #[serde(default)]
    pub addresses: Vec<IpAddr>,

    /// EmailAdresses that this Certificate is valid for
    #[serde(default)]
    pub emails: VString,

    /// The systemd services are reloaded
    #[serde(default)]
    pub reload_services: VString,

    /// The Systemd-Services have to be restarted to get the new certificates
    #[serde(default)]
    pub restart_services: VString,

    /// Commands that have to be run after the certificates have been issued if they don't have an systemd service
    #[serde(default)]
    pub trigger_commands: VString,

    /// The Algorithm for the Private Key
    #[serde(default)]
    pub algorithm: Algorithm,

    /// The Strength of the Private key.
    #[serde(default)]
    pub strength: Strength,

    #[serde(skip)]
    pub name: String,
}

pub async fn read_config<T: Default + DeserializeOwned>(mut file: File) -> T {
    let mut data = String::new();

    match file.read_to_string(&mut data).await {
        Ok(_) => {},
        Err(error) => {
            warn!("Failed to load config: {error}");

            return Default::default();
        },
    }

    match toml::from_str(&data) {
        Ok(output) => output,
        Err(error) => {
            warn!("Failed to parse toml file: {error}");

            Default::default()
        },
    }
}
