//! Acme client that supports multiple CAs and configs for sites that can be seperate from the mainconfig
#![allow(dead_code)]
#![allow(clippy::clone_on_copy)]
#![allow(clippy::identity_op)]

pub(crate) mod config;
pub(crate) mod consts;
pub(crate) mod macros;
pub(crate) mod process;
pub(crate) mod structs;
pub(crate) mod utils;

use crate::{
    config::SiteConfig,
    consts::*,
    macros::match_error,
    structs::{
        Arguments,
        ProcessorArgs,
    },
    utils::prefix_emails,
};
use acme2_eab::{
    Account,
    AccountBuilder,
    Directory,
    DirectoryBuilder,
};
use async_scoped::TokioScope;
use clap::Parser;
use config::CA;
use env_logger::init as log_init;
use libsystemd::daemon;
use log::*;
use openssl::{
    self,
    pkey::{
        PKey,
        Private,
    },
};
use process::process_site;
use reqwest::{
    Client,
    tls::Version,
};
use std::{
    collections::{
        HashMap,
        HashSet,
    },
    path::{
        Path,
        PathBuf,
    },
    str::FromStr,
    sync::Arc,
    time::Duration,
};
use tokio::{
    fs::{
        create_dir_all,
        read_dir,
    },
    io::{
        AsyncReadExt,
        AsyncWriteExt,
    },
    sync::Mutex,
};
use tokio_stream::{
    StreamExt,
    wrappers::ReadDirStream,
};
use zbus_systemd::systemd1;

type SafeSet<T> = Mutex<HashSet<T>>;


fn default_client() -> reqwest::Client {
    reqwest::Client::builder().min_tls_version(Version::TLS_1_2).https_only(true).pool_max_idle_per_host(POOL_SIZE).build().unwrap()
}

async fn load_privkey(path: PathBuf) -> Result<PKey<Private>, ()> {
    let mut file = match_error!(FILE_MODE.open(path).await=>Err(error)-> "Failed to open Private Key: {error}", Err(()));
    let mut data = String::new();
    if let Err(error) = file.read_to_string(&mut data).await {
        error!("Failed to read data for the key: {error}");
        return Err(());
    }
    match PKey::private_key_from_pem(data.as_bytes()) {
        Ok(key) => Ok(key),
        Err(error) => {
            error!("Failed to parse pem data: {error}");
            Err(())
        },
    }
}

async fn process_accounts(
    name: &String,
    ca: &CA,
    directories: &mut HashMap<String, Arc<Directory>>,
    accounts: &mut HashMap<String, Arc<Account>>,
    client: &Client,
    accountpath: PathBuf,
) {
    let directory = match directories.get(&ca.directory) {
        Some(directory) => directory.to_owned(),
        None => {
            match DirectoryBuilder::new(ca.directory.clone()).http_client(client.clone()).build().await {
                Ok(dir) => {
                    directories.insert(ca.directory.clone(), Arc::clone(&dir));
                    dir
                },
                Err(error) => {
                    error!("Failed to initialize directory for ca {name}: {error}");
                    return;
                },
            }
        },
    };
    let mut ac = AccountBuilder::new(Arc::clone(&directory));
    match ca.email_addresses.clone() {
        Some(addr) => {
            ac.contact(prefix_emails(addr));
        },
        None => {
            ac.contact(Vec::new());
            debug!("No Email address given")
        },
    }
    let accountkey = accountpath.join("file.pem").with_file_name(name.clone());
    let mut accountkeyfile = None;
    if accountkey.exists() {
        if let Ok(key) = load_privkey(accountkey).await {
            ac.private_key(key);
        }
    } else {
        info!("Registering for the CA {}", name.clone());
        accountkeyfile = match FILE_MODE_WRITE.open(accountkey).await {
            Ok(file) => Some(file),
            Err(error) => {
                error!("Failed to open the file for the accountkey: {error}");
                return;
            },
        }
    }
    if let Some(meta) = &directory.meta {
        // Collecting the errors about the metadata before annoying the admin about errors at different stages
        let mut errors = false;
        if let Some(tos) = &meta.terms_of_service {
            if !ca.tos_accepted {
                error!("Terms of Services were not agreed into: {tos}");
                errors = true;
            } else {
                ac.terms_of_service_agreed(true);
            }
        }
        if meta.external_account_required.unwrap_or(false) {
            if let Some(eab) = &ca.eab {
                match eab.key() {
                    Ok(private) => {
                        trace!("EAB Key info: Type={:?} Bits={}, Security-Bits={}", private.id(), private.bits(), private.security_bits());
                        ac.external_account_binding(eab.token.clone(), private);
                    },
                    Err(error) => {
                        error!("{error}");
                        errors = true;
                    },
                }
            } else {
                error!("eab_token and/or eab_key are unset, but the CA requires those.");
                errors = true;
            }
        } else if ca.eab.is_some() {
            warn!("The CA doesn't need EAB Tokens but they were configured")
        }
        if errors {
            return;
        }
    }
    let account = match ac.build().await {
        Ok(account) => {
            accounts.insert(name.clone(), Arc::clone(&account));
            account
        },
        Err(error) => {
            error!("Failed to get/create account: {error}");
            return;
        },
    };
    if let Some(mut keyfile) = accountkeyfile {
        let keydata = match_error!(account.private_key().private_key_to_pem_pkcs8()=>Err(error)-> "Failed to convert the private key to an pem: {error}");
        if let Err(error) = keyfile.write(keydata.as_slice()).await {
            error!("Failed to write the accountkey: {error}");
        }
    }
}

async fn racme(flags: Arguments) {
    let client = default_client();
    let systemd_access = daemon::booted();
    let mainconfig =
        config::read_config::<config::GeneralConfig>(match_error!(FILE_MODE.open(flags.config).await=>Err(error)-> "error reading the config: {error}")).await;
    trace!("Parsed Config: {mainconfig:?}");
    let files = ReadDirStream::new(match_error!(read_dir(mainconfig.sites_path.clone()).await=>Err(error)-> "could not read files from sites dir: {error}"));
    let mut siteconfigs = Vec::new();
    for file in files.filter(Result::is_ok).map(|file| file.unwrap().path()).collect::<Vec<PathBuf>>().await {
        let mut site = config::read_config::<SiteConfig>(FILE_MODE.open(file.clone()).await.unwrap()).await;
        site.name = file.file_stem().unwrap().to_str().unwrap().to_string();
        siteconfigs.push(site);
    }
    let used = siteconfigs.iter().map(|s| s.ca.clone()).collect::<HashSet<_>>();
    debug!("Used CAs: {used:?}");

    let mut directories = HashMap::<String, Arc<Directory>>::new();

    let mut accounts = HashMap::new();
    let accountpath = Path::new(&mainconfig.accounts_path).to_path_buf();
    if let Err(error) = create_dir_all(accountpath.clone()).await {
        error!("Failed to create the directory for the accounts: {error}");
        return;
    }
    for (name, ca) in mainconfig.ca.iter().filter(|(name, _)| used.contains(name.to_owned())) {
        process_accounts(name, ca, &mut directories, &mut accounts, &client, accountpath.clone()).await;
    }


    let restart_services = Mutex::new(HashSet::<String>::new());
    let reload_services = Mutex::new(HashSet::<String>::new());
    let certs = Path::new(&mainconfig.certificates_path).to_path_buf();

    if !certs.exists() {
        if let Err(error) = create_dir_all(certs.clone()).await {
            error!("Failed to create directory for all the certificates: {error}");
            return;
        }
    }
    let challengepath = mainconfig.http_challenge_path.and_then(|path| PathBuf::from_str(path.as_str()).ok());
    let dnsserver = None;

    unsafe {
        TokioScope::scope_and_collect(|scope| {
            for site in siteconfigs {
                if let Some(account) = accounts.get(&site.ca) {
                    scope.spawn(process_site(ProcessorArgs::new(
                        site,
                        Arc::clone(account),
                        &reload_services,
                        &restart_services,
                        certs.clone(),
                        challengepath.clone(),
                        dnsserver.clone(),
                    )));
                } else {
                    error!("Could not process site {} because of previous errors", site.name)
                }
            }
        })
    }
    .await;

    if systemd_access {
        let conn = match_error!(zbus_systemd::zbus::Connection::system().await=>Err(error)-> "Failed to connect with the systemd manager: {error}");

        let systemd_manager = systemd1::ManagerProxy::new(&conn).await.unwrap();
        let restart_services = restart_services.into_inner();

        for service in reload_services.into_inner().difference(&restart_services.clone()) {
            match systemd_manager.reload_unit(service.to_owned(), "replace".to_string()).await {
                Ok(_) => info!("Reloaded {service}"),
                Err(error) => error!("Failed to reload service {service}: {error}"),
            };
        }

        for service in restart_services.iter() {
            match systemd_manager.restart_unit(service.to_owned(), "replace".to_string()).await {
                Ok(_) => info!("Restarted {service}"),
                Err(error) => error!("Failed to restart service {service}: {error}"),
            };
        }
    }
}

fn main() {
    log_init();
    let runtime = tokio::runtime::Builder::new_current_thread().enable_all().build().unwrap();
    runtime.block_on(racme(Arguments::parse()));
    runtime.shutdown_timeout(Duration::from_secs(1));
}
