use std::{
    fs::Permissions,
    os::unix::fs::PermissionsExt,
    path::PathBuf,
};

use crate::{
    config::{
        Algorithm,
        match_algo,
    },
    consts::{
        ATTEMPTS,
        FILE_MODE,
        FILE_MODE_WRITE,
        WAIT_TIME,
    },
    load_privkey,
    match_error,
    structs::{
        ProcessorArgs,
        San,
    },
    utils::{
        gen_key,
        is_matching,
    },
};
use acme2_eab::{
    Authorization,
    ChallengeStatus,
    Csr,
    Identifier,
    OrderBuilder,
    OrderStatus,
};
use async_scoped::TokioScope;
use log::*;
use openssl::{
    hash::MessageDigest,
    stack::Stack,
    x509::{
        X509,
        X509Extension,
        X509Req,
        X509v3Context,
        extension::{
            BasicConstraints,
            ExtendedKeyUsage,
            KeyUsage,
            SubjectAlternativeName,
        },
    },
};
use tokio::{
    fs::{
        create_dir_all,
        remove_file,
    },
    io::{
        AsyncReadExt,
        AsyncWriteExt,
    },
};

fn gen_stack(args: &ProcessorArgs, context: X509v3Context) -> Stack<X509Extension> {
    let mut stack = Stack::new().unwrap();

    let _ = stack.push({
        let mut subaltname = SubjectAlternativeName::new();
        for san in args.san() {
            match san {
                San::Dns(domain) => subaltname.dns(&domain),
                San::Email(email) => subaltname.email(&email),
                San::IPAddress(ip) => subaltname.ip(ip.to_canonical().to_string().as_ref()),
            };
        }
        subaltname.build(&context).unwrap()
    });
    stack.push(BasicConstraints::new().critical().build().unwrap()).unwrap();
    stack.push(ExtendedKeyUsage::new().server_auth().critical().build().unwrap()).unwrap();
    stack.push(KeyUsage::new().critical().digital_signature().key_agreement().key_encipherment().build().unwrap()).unwrap();
    stack
}


pub async fn process_site(args: ProcessorArgs<'_>) {
    let mut cert_renew = false;
    info!("Processing Site {}", args.name());
    let directory = args.certificate_dir().join(args.name().clone());
    if !directory.exists() {
        if let Err(error) = create_dir_all(directory.clone()).await {
            error!("Failed to create directory for site {}: {}", args.name(), error);
            return;
        };
        cert_renew = true;
    }
    let private_key_file = directory.join("privkey.pem");
    let mut private_key;
    let mut write_pkey = false;
    if !private_key_file.exists() {
        cert_renew = true;
        write_pkey = true;
        private_key = match_error!(gen_key(args.algorithm(), args.strength())=>Err(error)-> "Aborting processing the site due to problem with the certificate generation: {error}");
    } else if let Ok(key) = load_privkey(private_key_file.clone()).await {
        private_key = key;
        if !match_algo(&private_key, args.algorithm(), args.strength()) {
            info!("Algorithm for the private key has changed, updating the key");
            cert_renew = true;
            write_pkey = true;
            private_key = match_error!(gen_key(args.algorithm(), args.strength())=>Err(error)-> "Aborting processing the site due to problem with the certificate generation: {error}");
        }
    } else {
        error!("Failed to parse the private key. Renewing the private key.");
        write_pkey = true;
        cert_renew = true;
        private_key = match_error!(gen_key(args.algorithm(), args.strength())=>Err(error)-> "Aborting processing the site due to problem with the certificate generation: {error}");
    }
    if write_pkey {
        let pkey = private_key.private_key_to_pem_pkcs8().unwrap();
        let mut file = match_error!(FILE_MODE_WRITE.open(private_key_file.clone()).await=>Err(error)-> "Failed to write new private key: {error}");
        match_error!(file.write_all(&pkey).await=>Err(error)->"Failed to write new private key: {error}");
    }
    let pubkey_filename = directory.join("pubkey.pem");
    if pubkey_filename.exists() {
        let mut file = match_error!(FILE_MODE.open(pubkey_filename.clone()).await=>Err(error)->  "Failed to open publickey. Aborting processing: {error}");
        let mut data = String::new();
        if let Err(error) = file.read_to_string(&mut data).await {
            cert_renew = true;
            error!("Failed to read public key: {error}")
        } else {
            let pubkey = match X509::from_pem(data.as_bytes()) {
                Ok(key) => key,
                Err(_) => todo!(),
            };
            if !is_matching(pubkey, args.refresh_time(), args.san()) {
                info!("Subject Alternative Names differ from Certifcate");
                cert_renew = true;
            };
        }
    } else {
        cert_renew = true;
    }
    if !cert_renew {
        info!("Site {} doesn't need an update for the certificate.", args.name());
        return;
    }
    info!("Renewing Certificate for site {}", args.name());
    let mut builder = OrderBuilder::new(args.account());
    builder.set_identifiers(args.san().iter().map(|s| s.to_owned().into()).collect::<Vec<Identifier>>());
    let mut order = match_error!(builder.build().await=>Err(error)->  "Failed order the certificate: {error}");
    let authorizations = match_error!(order.authorizations().await=>Err(error)->  "Failed to get the authorizations: {error}");
    let (_, result) = tokio::join! {
        unsafe {
            TokioScope::scope_and_collect(|scope|{
                for auth in authorizations {
                    scope.spawn(process_auth(auth, args.challenge_dir(), args.dnsserver()));
                }
            })
        },
        order.wait_ready(WAIT_TIME, ATTEMPTS),
    };
    order = match_error!(result=>Err(error)->  "Failed to process order: {error}");
    if order.status == OrderStatus::Invalid {
        error!("Failed the Order, check the logs for more information");
        return;
    }
    let mut csr = X509Req::builder().unwrap();
    if let Err(error) = csr.set_pubkey(&private_key) {
        error!("failed to add the public key: {error}");
        return;
    }
    let _ = csr.add_extensions(&gen_stack(&args, csr.x509v3_context(None)));
    if let Err(error) = csr.sign(
        &private_key,
        match args.algorithm() {
            Algorithm::Rsa => MessageDigest::sha3_512(),
            _ => MessageDigest::null(),
        },
    ) {
        error!("Failed to sign Request: {error}");
        return;
    }
    order = match_error!(order.finalize(Csr::Custom(csr.build())).await=>Err(error)->  "Failed to finalize the order: {error}");
    order = match_error!(order.wait_done(WAIT_TIME, ATTEMPTS).await=>Err(error)->  "Failed to finalize the order: {error}");
    if order.status != OrderStatus::Valid {
        error!("Failed to complete the order: check the logs for more information");
        return;
    }
    let certs = order.certificate().await.unwrap().unwrap();
    debug!("Received {} certificates.", certs.len());
    let mut pubkey_file = match_error!(FILE_MODE_WRITE.open(pubkey_filename).await=>Err(error)->  "Failed to open the file for the publickey: {error}");
    match_error!(pubkey_file.write_all(&certs[0].to_pem().unwrap()).await=>Err(error)->  "Failed to write the publickey: {error}");
    let mut fullchain = match_error!(FILE_MODE_WRITE.open(directory.join("fullchain.pem")).await=>Err(error)->  "failed to open the fullchain.pem: {error}");
    for cert in certs.clone() {
        let _ = fullchain.write_all(&cert.to_pem().unwrap()).await;
    }
    let mut bundle = match_error!(FILE_MODE_WRITE.open(directory.join("bundle.pem")).await=>Err(error)->  "failed to open the bundle.pem: {error}");
    let _ = bundle.write_all(&private_key.private_key_to_pem_pkcs8().unwrap()).await;
    let _ = bundle.write_all(&certs[0].to_pem().unwrap()).await;
    info!("Processing of {} successful", args.name());
    let mut services = args.reload_list().await;
    for service in &args.reload_services() {
        services.insert(service.to_owned());
    }

    let mut services = args.restart_list().await;
    for service in &args.restart_services() {
        services.insert(service.to_owned());
    }
}

pub async fn process_auth(auth: Authorization, challenge_dir: Option<PathBuf>, dnsserver: Option<()>) {
    if let Some(_dnschallenge) = auth.get_challenge("dns-01") {
        if let Some(_dnsserver) = dnsserver {
        } else {
            debug!("DNS-01 is disabled")
        }
    }
    if !auth.wildcard.unwrap_or(false) {
        if let Some(mut challenge) = auth.get_challenge("http-01") {
            trace!("CA has an http-challenge");
            if let Some(directory) = challenge_dir {
                match_error!(create_dir_all(directory.clone()).await=>Err(error)->  "Failed to ensure the directory exists: {error}");
                let filename = directory.join(challenge.token.clone().unwrap());
                let mut challengefile =
                    match_error!(FILE_MODE_WRITE.open(filename.clone()).await=>Err(error)->  "Failed to open the file for the http-challenge: {error}");
                match_error!(challengefile.set_permissions(Permissions::from_mode(0o644)).await=>Err(error)->  "Failed to give the file the nessesary permissions: {error}");
                match_error!(
                    challengefile.write_all(challenge.key_authorization().unwrap().unwrap().as_bytes()).await=>Err(error)->
                    "Failed to write the challenge to the file: {error}"
                );
                drop(challengefile);
                challenge = match_error!(challenge.validate().await=>Err(error)->  "failed to inform the server about the deployed challenge: {error}");
                challenge = match_error!(challenge.wait_done(WAIT_TIME, ATTEMPTS).await=>Err(error)->  "Server took too long for the validation: {error}");
                if challenge.status == ChallengeStatus::Invalid {
                    error!("Failed the challenge: {}", challenge.error.unwrap())
                } else {
                    info!("Challenge for {} successful", auth.identifier.value)
                }
                if let Err(error) = remove_file(filename).await {
                    error!("Failed to remove the challenge after it has been deployed: {error}");
                }
                return;
            } else {
                debug!("No Challengedir given")
            }
        }
    }
    error!("Cannot prove the challenges: {}", auth.challenges.iter().map(|c| c.r#type.clone()).collect::<Vec<_>>().join(", "))
}
