use std::{
    collections::HashSet,
    fmt::Display,
    net::IpAddr,
    path::PathBuf,
    sync::Arc,
};

use acme2_eab::{
    Account,
    Identifier,
};
use clap::Parser;
use macro_rules_attribute::macro_rules_derive;
use openssl::x509::GeneralName;
use tokio::sync::MutexGuard;

use crate::{
    SafeSet,
    config::{
        Algorithm,
        SiteConfig,
        Strength,
    },
    macros::{
        DefDer,
        Hashable,
        attr_function,
    },
};


#[macro_rules_derive(DefDer)]
#[derive(Parser)]
pub struct Arguments {
    pub config: String,
}

#[macro_rules_derive(DefDer)]
pub struct ProcessorArgs<'a> {
    site: SiteConfig,
    account: Arc<Account>,
    reload_services: &'a SafeSet<String>,
    restart_services: &'a SafeSet<String>,
    certificate_dir: PathBuf,
    refresh_time: u32,
    challenge_dir: Option<PathBuf>,
    dnsserver: Option<()>,
}

impl<'a: 'b, 'b> ProcessorArgs<'a> {
    attr_function!(pub name site => String);

    attr_function!(pub algorithm site => Algorithm);

    attr_function!(pub strength site => Strength);

    attr_function!(pub domains site => Vec<String>);

    attr_function!(pub addresses site => Vec<IpAddr>);

    attr_function!(pub emails site => Vec<String>);

    attr_function!(pub restart_services site => Vec<String>);

    attr_function!(pub reload_services site => Vec<String>);

    attr_function!(pub certificate_dir => PathBuf);

    attr_function!(pub refresh_time => u32);

    attr_function!(pub challenge_dir => Option<PathBuf>);

    attr_function!(pub dnsserver => Option<()>);

    pub fn new(
        site: SiteConfig,
        account: Arc<Account>,
        reload_services: &'a SafeSet<String>,
        restart_services: &'a SafeSet<String>,
        certificate_dir: PathBuf,
        http_challenge_dir: Option<PathBuf>,
        dnsserver: Option<()>,
    ) -> Self {
        ProcessorArgs {
            site,
            account,
            reload_services,
            restart_services,
            certificate_dir,
            refresh_time: 7,
            challenge_dir: http_challenge_dir,
            dnsserver,
        }
    }

    pub fn account(&self) -> Arc<Account> {
        Arc::clone(&self.account)
    }

    pub async fn restart_list(&self) -> MutexGuard<'b, HashSet<String>> {
        self.restart_services.lock().await
    }

    pub async fn reload_list(&self) -> MutexGuard<'b, HashSet<String>> {
        self.reload_services.lock().await
    }

    pub fn san(&self) -> Vec<San> {
        let mut sans = Vec::with_capacity(self.site.domains.len() + self.site.emails.len() + self.site.addresses.len());
        for domain in self.domains() {
            sans.push(San::Dns(domain));
        }
        for address in self.addresses() {
            sans.push(San::IPAddress(address));
        }
        for email in self.emails() {
            sans.push(San::Email(email));
        }
        sans
    }
}

#[macro_rules_derive(DefDer, Hashable)]
#[derive(PartialEq)]
pub enum San {
    Dns(String),
    Email(String),
    IPAddress(IpAddr),
}

impl From<GeneralName> for San {
    fn from(value: GeneralName) -> Self {
        if let Some(dns) = value.dnsname() {
            return Self::Dns(dns.to_owned());
        }
        if let Some(ipaddr) = value.ipaddress() {
            if ipaddr.len() == 4 {
                let mut addr = [0u8; 4];
                addr.copy_from_slice(ipaddr);
                return Self::IPAddress(IpAddr::from(addr));
            } else {
                let mut addr = [0u8; 16];
                addr.copy_from_slice(ipaddr);
                return Self::IPAddress(IpAddr::from(addr));
            }
        }
        if let Some(email) = value.email() {
            return Self::Email(email.to_owned());
        }
        unreachable!();
    }
}

impl From<San> for Identifier {
    fn from(value: San) -> Self {
        match value {
            San::Dns(domain) => {
                Identifier {
                    r#type: "dns".into(),
                    value: domain,
                }
            },
            San::Email(email) => {
                Identifier {
                    r#type: "email".into(),
                    value: email,
                }
            },
            San::IPAddress(ip) => {
                Identifier {
                    r#type: "ip".into(),
                    value: ip.to_string(),
                }
            },
        }
    }
}

#[macro_rules_derive(DefDer)]
pub struct Error(String);

impl Error {
    #[inline]
    pub fn err<T>(message: String) -> Result<T, Self> {
        Err(Self::new(message))
    }

    #[inline]
    pub fn new(message: String) -> Self {
        Self(message)
    }
}

impl Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(&self.0)
    }
}
